@@ -33,10 +33,12 @@ extern "C" {
  */
 
 enum connman_service_type {
-	CONNMAN_SERVICE_TYPE_UNKNOWN  = 0,
-	CONNMAN_SERVICE_TYPE_ETHERNET = 1,
-	CONNMAN_SERVICE_TYPE_WIFI     = 2,
-	CONNMAN_SERVICE_TYPE_WIMAX    = 3,
+	CONNMAN_SERVICE_TYPE_UNKNOWN   = 0,
+	CONNMAN_SERVICE_TYPE_ETHERNET  = 1,
+	CONNMAN_SERVICE_TYPE_WIFI      = 2,
+	CONNMAN_SERVICE_TYPE_WIMAX     = 3,
+	CONNMAN_SERVICE_TYPE_BLUETOOTH = 4,
+	CONNMAN_SERVICE_TYPE_CELLULAR  = 5,
 };
 
 enum connman_service_mode {
@@ -50,7 +52,7 @@ enum connman_service_security {
 	CONNMAN_SERVICE_SECURITY_NONE    = 1,
 	CONNMAN_SERVICE_SECURITY_WEP     = 2,
 	CONNMAN_SERVICE_SECURITY_WPA     = 3,
-	CONNMAN_SERVICE_SECURITY_WPA2    = 4,
+	CONNMAN_SERVICE_SECURITY_RSN     = 4,
 };
 
 enum connman_service_state {
@@ -64,13 +66,18 @@ enum connman_service_state {
 	CONNMAN_SERVICE_STATE_FAILURE       = 7,
 };
 
+enum connman_service_error {
+	CONNMAN_SERVICE_ERROR_UNKNOWN     = 0,
+	CONNMAN_SERVICE_ERROR_DHCP_FAILED = 1,
+};
+
 struct connman_service;
 
-extern struct connman_service *connman_service_create(void);
-extern struct connman_service *connman_service_ref(struct connman_service *service);
-extern void connman_service_unref(struct connman_service *service);
+struct connman_service *connman_service_create(void);
+struct connman_service *connman_service_ref(struct connman_service *service);
+void connman_service_unref(struct connman_service *service);
 
-extern int connman_service_set_favorite(struct connman_service *service,
+int connman_service_set_favorite(struct connman_service *service,
 						connman_bool_t favorite);
 
 #ifdef __cplusplus
