@@ -33,10 +33,11 @@
 
 #define CONNMAN_API_SUBJECT_TO_CHANGE
 #include <connman/device.h>
+#include <connman/option.h>
+#include <connman/inet.h>
 #include <connman/dbus.h>
 #include <connman/log.h>
 
-#include "inet.h"
 #include "supplicant.h"
 
 #define TIMEOUT 5000
@@ -149,7 +150,8 @@ enum supplicant_state {
 struct supplicant_result {
 	char *path;
 	char *name;
-	char *addr;
+	unsigned char *addr;
+	unsigned int addr_len;
 	unsigned char *ssid;
 	unsigned int ssid_len;
 	dbus_uint16_t capabilities;
@@ -158,6 +160,7 @@ struct supplicant_result {
 	gboolean has_wpa;
 	gboolean has_rsn;
 	gboolean has_wps;
+	dbus_int32_t frequency;
 	dbus_int32_t quality;
 	dbus_int32_t noise;
 	dbus_int32_t level;
@@ -211,7 +214,7 @@ static struct supplicant_task *find_task_by_path(const char *path)
 	for (list = task_list; list; list = list->next) {
 		struct supplicant_task *task = list->data;
 
-		if (g_str_equal(task->path, path) == TRUE)
+		if (g_strcmp0(task->path, path) == 0)
 			return task;
 	}
 
@@ -232,7 +235,7 @@ static void add_interface_reply(DBusPendingCall *call, void *user_data)
 		return;
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
-		goto done;
+		goto failed;
 
 	dbus_error_init(&error);
 
@@ -243,7 +246,7 @@ static void add_interface_reply(DBusPendingCall *call, void *user_data)
 			dbus_error_free(&error);
 		} else
 			connman_error("Wrong arguments for add interface");
-		goto done;
+		goto failed;
 	}
 
 	DBG("path %s", path);
@@ -253,13 +256,23 @@ static void add_interface_reply(DBusPendingCall *call, void *user_data)
 
 	connman_device_set_powered(task->device, TRUE);
 
-done:
 	dbus_message_unref(reply);
+
+	return;
+
+failed:
+	task_list = g_slist_remove(task_list, task);
+
+	connman_device_unref(task->device);
+
+	free_task(task);
 }
 
 static int add_interface(struct supplicant_task *task)
 {
+	const char *driver = connman_option_get_string("wifi");
 	DBusMessage *message;
+	DBusMessageIter array, dict;
 	DBusPendingCall *call;
 
 	DBG("task %p", task);
@@ -269,8 +282,20 @@ static int add_interface(struct supplicant_task *task)
 	if (message == NULL)
 		return -ENOMEM;
 
-	dbus_message_append_args(message, DBUS_TYPE_STRING, &task->ifname,
-							DBUS_TYPE_INVALID);
+	dbus_message_iter_init_append(message, &array);
+
+	dbus_message_iter_append_basic(&array,
+					DBUS_TYPE_STRING, &task->ifname);
+
+	dbus_message_iter_open_container(&array, DBUS_TYPE_ARRAY,
+			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
+			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
+			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
+
+	connman_dbus_dict_append_variant(&dict, "driver",
+						DBUS_TYPE_STRING, &driver);
+
+	dbus_message_iter_close_container(&array, &dict);
 
 	if (dbus_connection_send_with_reply(connection, message,
 						&call, TIMEOUT) == FALSE) {
@@ -381,7 +406,7 @@ static void remove_interface_reply(DBusPendingCall *call, void *user_data)
 
 	connman_device_unref(task->device);
 
-	inet_ifdown(task->ifindex);
+	connman_inet_ifdown(task->ifindex);
 
 	free_task(task);
 
@@ -712,7 +737,7 @@ static int set_network(struct supplicant_task *task,
 					DBUS_TYPE_BYTE, &network, len);
 
 	if (g_ascii_strcasecmp(security, "wpa") == 0 ||
-				g_ascii_strcasecmp(security, "wpa2") == 0) {
+				g_ascii_strcasecmp(security, "rsn") == 0) {
 		const char *key_mgmt = "WPA-PSK";
 		connman_dbus_dict_append_variant(&dict, "key_mgmt",
 						DBUS_TYPE_STRING, &key_mgmt);
@@ -803,22 +828,61 @@ static int initiate_scan(struct supplicant_task *task)
 	return 0;
 }
 
-static char *build_group(const unsigned char *ssid, unsigned int ssid_len,
+static struct {
+	char *name;
+	char *value;
+} special_ssid[] = {
+	{ "<hidden>", "hidden"  },
+	{ "default",  "linksys" },
+	{ "wireless"  },
+	{ "linksys"   },
+	{ "netgear"   },
+	{ "dlink"     },
+	{ "2wire"     },
+	{ "compaq"    },
+	{ "tsunami"   },
+	{ "comcomcom", "3com"     },
+	{ "3Com",      "3com"     },
+	{ "Symbol",    "symbol"   },
+	{ "Motorola",  "motorola" },
+	{ "Wireless" , "wireless" },
+	{ "WLAN",      "wlan"     },
+	{ }
+};
+
+static char *build_group(const char *addr, const char *name,
+			const unsigned char *ssid, unsigned int ssid_len,
 					const char *mode, const char *security)
 {
 	GString *str;
 	unsigned int i;
 
-	if (ssid_len < 1)
+	if (addr == NULL)
 		return NULL;
 
 	str = g_string_sized_new((ssid_len * 2) + 24);
 	if (str == NULL)
 		return NULL;
 
-	for (i = 0; i < ssid_len; i++)
-		g_string_append_printf(str, "%02x", ssid[i]);
+	for (i = 0; special_ssid[i].name; i++) {
+		if (g_strcmp0(special_ssid[i].name, name) == 0) {
+			if (special_ssid[i].value == NULL)
+				g_string_append_printf(str, "%s_%s",
+								name, addr);
+			else
+				g_string_append_printf(str, "%s_%s",
+						special_ssid[i].value, addr);
+			goto done;
+		}
+	}
 
+	if (ssid_len > 0 && ssid[0] != '\0') {
+		for (i = 0; i < ssid_len; i++)
+			g_string_append_printf(str, "%02x", ssid[i]);
+	} else
+		g_string_append_printf(str, "hidden_%s", addr);
+
+done:
 	g_string_append_printf(str, "_%s_%s", mode, security);
 
 	return g_string_free(str, FALSE);
@@ -838,25 +902,20 @@ static void extract_addr(DBusMessageIter *value,
 	if (addr_len != 6)
 		return;
 
-	eth = (void *) addr;
-
-	result->addr = g_try_malloc0(18);
+	result->addr = g_try_malloc(addr_len);
 	if (result->addr == NULL)
 		return;
 
-	snprintf(result->addr, 18, "%02X:%02X:%02X:%02X:%02X:%02X",
-						eth->ether_addr_octet[0],
-						eth->ether_addr_octet[1],
-						eth->ether_addr_octet[2],
-						eth->ether_addr_octet[3],
-						eth->ether_addr_octet[4],
-						eth->ether_addr_octet[5]);
+	memcpy(result->addr, addr, addr_len);
+	result->addr_len = addr_len;
 
-	result->path = g_try_malloc0(18);
+	result->path = g_try_malloc0(13);
 	if (result->path == NULL)
 		return;
 
-	snprintf(result->path, 18, "%02X_%02X_%02X_%02X_%02X_%02X",
+	eth = (void *) addr;
+
+	snprintf(result->path, 13, "%02x%02x%02x%02x%02x%02x",
 						eth->ether_addr_octet[0],
 						eth->ether_addr_octet[1],
 						eth->ether_addr_octet[2],
@@ -878,9 +937,6 @@ static void extract_ssid(DBusMessageIter *value,
 	if (ssid_len < 1)
 		return;
 
-	if (ssid[0] == '\0')
-		return;
-
 	result->ssid = g_try_malloc(ssid_len);
 	if (result->ssid == NULL)
 		return;
@@ -951,6 +1007,33 @@ static void extract_capabilites(DBusMessageIter *value,
 		result->has_wep = TRUE;
 }
 
+static unsigned char calculate_strength(struct supplicant_result *result)
+{
+	if (result->quality < 0) {
+		unsigned char strength;
+
+		if (result->level > 0)
+			strength = 100 - result->level;
+		else
+			strength = 120 + result->level;
+
+		if (strength > 100)
+			strength = 100;
+
+		return strength;
+	}
+
+	return result->quality;
+}
+
+static unsigned short calculate_channel(struct supplicant_result *result)
+{
+	if (result->frequency < 0)
+		return 0;
+
+	return (result->frequency - 2407) / 5;
+}
+
 static void get_properties(struct supplicant_task *task);
 
 static void properties_reply(DBusPendingCall *call, void *user_data)
@@ -961,6 +1044,7 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	DBusMessage *reply;
 	DBusMessageIter array, dict;
 	unsigned char strength;
+	unsigned short channel, frequency;
 	const char *mode, *security;
 	char *group;
 
@@ -979,6 +1063,10 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	}
 
 	memset(&result, 0, sizeof(result));
+	result.frequency = -1;
+	result.quality = -1;
+	result.level = 0;
+	result.noise = 0;
 
 	dbus_message_iter_init(reply, &array);
 
@@ -1024,6 +1112,8 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 			extract_wpsie(&value, &result);
 		else if (g_str_equal(key, "capabilities") == TRUE)
 			extract_capabilites(&value, &result);
+		else if (g_str_equal(key, "frequency") == TRUE)
+			dbus_message_iter_get_basic(&value, &result.frequency);
 		else if (g_str_equal(key, "quality") == TRUE)
 			dbus_message_iter_get_basic(&value, &result.quality);
 		else if (g_str_equal(key, "noise") == TRUE)
@@ -1042,10 +1132,18 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	if (result.path[0] == '\0')
 		goto done;
 
-	strength = result.quality;
+	if (result.frequency > 0 && result.frequency < 14)
+		result.frequency = 2407 + (5 * result.frequency);
+	else if (result.frequency == 14)
+		result.frequency = 2484;
+
+	strength = calculate_strength(&result);
+	channel  = calculate_channel(&result);
+
+	frequency = (result.frequency < 0) ? 0 : result.frequency;
 
 	if (result.has_rsn == TRUE)
-		security = "wpa2";
+		security = "rsn";
 	else if (result.has_wpa == TRUE)
 		security = "wpa";
 	else if (result.has_wep == TRUE)
@@ -1055,7 +1153,9 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 
 	mode = (result.adhoc == TRUE) ? "adhoc" : "managed";
 
-	group = build_group(result.ssid, result.ssid_len, mode, security);
+	group = build_group(result.path, result.name,
+					result.ssid, result.ssid_len,
+							mode, security);
 
 	network = connman_device_get_network(task->device, result.path);
 	if (network == NULL) {
@@ -1072,17 +1172,8 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 		connman_network_set_protocol(network,
 						CONNMAN_NETWORK_PROTOCOL_IP);
 
-		connman_network_set_string(network, "Address", result.addr);
-
-		connman_network_set_group(network, group);
-
-		if (result.name != NULL && result.name[0] != '\0')
-			connman_network_set_string(network, "Name", result.name);
-
-		connman_network_set_uint8(network, "Strength", strength);
-
-		connman_network_set_string(network, "WiFi.Mode", mode);
-		connman_network_set_string(network, "WiFi.Security", security);
+		connman_network_set_address(network, result.addr,
+							result.addr_len);
 
 		if (connman_device_add_network(task->device, network) < 0) {
 			connman_network_unref(network);
@@ -1090,12 +1181,8 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 		}
 	}
 
-	connman_network_set_group(network, group);
-
-	g_free(group);
-
 	if (result.name != NULL && result.name[0] != '\0')
-		connman_network_set_string(network, "Name", result.name);
+		connman_network_set_name(network, result.name);
 
 	connman_network_set_blob(network, "WiFi.SSID",
 						result.ssid, result.ssid_len);
@@ -1107,10 +1194,16 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 				(result.has_wps == TRUE) ? "WPS" : "no WPS");
 
 	connman_network_set_available(network, TRUE);
-	connman_network_set_uint8(network, "Strength", strength);
+	connman_network_set_strength(network, strength);
 
+	connman_network_set_uint16(network, "Frequency", frequency);
+	connman_network_set_uint16(network, "WiFi.Channel", channel);
 	connman_network_set_string(network, "WiFi.Security", security);
 
+	connman_network_set_group(network, group);
+
+	g_free(group);
+
 done:
 	g_free(result.path);
 	g_free(result.addr);
@@ -1395,7 +1488,7 @@ int supplicant_start(struct connman_device *device)
 		return -ENOMEM;
 
 	task->ifindex = connman_device_get_index(device);
-	task->ifname = inet_index2name(task->ifindex);
+	task->ifname = connman_inet_ifname(task->ifindex);
 
 	if (task->ifname == NULL) {
 		g_free(task);
