@@ -30,17 +30,17 @@
 
 #include "connman.h"
 
+static DBusConnection *connection = NULL;
+
 struct connman_device {
 	struct connman_element element;
 	enum connman_device_type type;
 	enum connman_device_mode mode;
-	enum connman_device_policy policy;
 	connman_bool_t secondary;
 	connman_bool_t powered;
 	connman_bool_t carrier;
 	connman_bool_t scanning;
 	connman_bool_t disconnected;
-	connman_uint8_t priority;
 	connman_uint16_t scan_interval;
 	char *name;
 	char *node;
@@ -49,6 +49,7 @@ struct connman_device {
 	char *ident;
 	unsigned int connections;
 	guint scan_timeout;
+	struct connman_ipconfig *ipconfig;
 
 	struct connman_device_driver *driver;
 	void *driver_data;
@@ -58,6 +59,9 @@ struct connman_device {
 	char *last_network;
 	struct connman_network *network;
 	GHashTable *networks;
+
+	DBusMessage *pending;
+	guint timeout;
 };
 
 static gboolean device_scan_trigger(gpointer user_data)
@@ -129,75 +133,31 @@ static const char *type2string(enum connman_device_type type)
 	return NULL;
 }
 
-static const char *policy2string(enum connman_device_policy policy)
+static int set_connected(struct connman_device *device,
+						connman_bool_t connected)
 {
-	switch (policy) {
-	case CONNMAN_DEVICE_POLICY_UNKNOWN:
-		break;
-	case CONNMAN_DEVICE_POLICY_IGNORE:
-		return "ignore";
-	case CONNMAN_DEVICE_POLICY_OFF:
-		return "off";
-	case CONNMAN_DEVICE_POLICY_AUTO:
-		return "auto";
-	case CONNMAN_DEVICE_POLICY_MANUAL:
-		return "manual";
-	}
-
-	return NULL;
-}
-
-static enum connman_device_policy string2policy(const char *policy)
-{
-	if (g_str_equal(policy, "ignore") == TRUE)
-		return CONNMAN_DEVICE_POLICY_IGNORE;
-	else if (g_str_equal(policy, "off") == TRUE)
-		return CONNMAN_DEVICE_POLICY_OFF;
-	else if (g_str_equal(policy, "auto") == TRUE)
-		return CONNMAN_DEVICE_POLICY_AUTO;
-	else if (g_str_equal(policy, "manual") == TRUE)
-		return CONNMAN_DEVICE_POLICY_MANUAL;
-	else
-		return CONNMAN_DEVICE_POLICY_UNKNOWN;
-}
-
-static int set_carrier(struct connman_device *device, connman_bool_t carrier)
-{
-	struct connman_service *service;
-
-	service = __connman_service_lookup_from_device(device);
-	__connman_service_set_carrier(service, carrier);
-
-	if (carrier == TRUE) {
+	if (connected == TRUE) {
 		enum connman_element_type type = CONNMAN_ELEMENT_TYPE_UNKNOWN;
 		struct connman_element *element;
 
 		device->disconnected = TRUE;
 
-		switch (device->policy) {
-		case CONNMAN_DEVICE_POLICY_UNKNOWN:
-		case CONNMAN_DEVICE_POLICY_IGNORE:
-		case CONNMAN_DEVICE_POLICY_OFF:
-		case CONNMAN_DEVICE_POLICY_MANUAL:
-			return 0;
-		case CONNMAN_DEVICE_POLICY_AUTO:
-			break;
-		}
-
 		switch (device->element.ipv4.method) {
-		case CONNMAN_IPV4_METHOD_UNKNOWN:
-		case CONNMAN_IPV4_METHOD_OFF:
+		case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
+		case CONNMAN_IPCONFIG_METHOD_OFF:
 			return 0;
-		case CONNMAN_IPV4_METHOD_STATIC:
+		case CONNMAN_IPCONFIG_METHOD_STATIC:
 			type = CONNMAN_ELEMENT_TYPE_IPV4;
 			break;
-		case CONNMAN_IPV4_METHOD_DHCP:
+		case CONNMAN_IPCONFIG_METHOD_DHCP:
 			type = CONNMAN_ELEMENT_TYPE_DHCP;
 			break;
 		}
 
 		element = connman_element_create(NULL);
 		if (element != NULL) {
+			struct connman_service *service;
+
 			element->type  = type;
 			element->index = device->element.index;
 
@@ -207,15 +167,29 @@ static int set_carrier(struct connman_device *device, connman_bool_t carrier)
 
 			device->disconnected = FALSE;
 
+			service = __connman_service_lookup_from_device(device);
 			__connman_service_indicate_state(service,
 					CONNMAN_SERVICE_STATE_CONFIGURATION);
 		}
-	} else
+	} else {
 		connman_element_unregister_children(&device->element);
 
+		device->disconnected = TRUE;
+	}
+
 	return 0;
 }
 
+static int set_carrier(struct connman_device *device, connman_bool_t carrier)
+{
+	struct connman_service *service;
+
+	service = __connman_service_lookup_from_device(device);
+	__connman_service_set_carrier(service, carrier);
+
+	return set_connected(device, carrier);
+}
+
 static int set_powered(struct connman_device *device, connman_bool_t powered)
 {
 	struct connman_device_driver *driver = device->driver;
@@ -247,64 +221,6 @@ static int set_powered(struct connman_device *device, connman_bool_t powered)
 	return err;
 }
 
-static int set_policy(DBusConnection *connection,
-				struct connman_device *device,
-					enum connman_device_policy policy)
-{
-	DBusMessage *signal;
-	DBusMessageIter entry, value;
-	const char *str, *key = "Policy";
-	int err = 0;
-
-	DBG("device %p policy %d", device, policy);
-
-	if (device->policy == policy)
-		return 0;
-
-	switch (policy) {
-	case CONNMAN_DEVICE_POLICY_UNKNOWN:
-		return -EINVAL;
-	case CONNMAN_DEVICE_POLICY_IGNORE:
-		break;
-	case CONNMAN_DEVICE_POLICY_OFF:
-		if (device->powered == TRUE)
-			err = set_powered(device, FALSE);
-		break;
-	case CONNMAN_DEVICE_POLICY_AUTO:
-	case CONNMAN_DEVICE_POLICY_MANUAL:
-		if (device->powered == FALSE)
-			err = set_powered(device, TRUE);
-		else
-			err = set_carrier(device, device->carrier);
-		break;
-	}
-
-	if (err < 0)
-		return err;
-
-	device->policy = policy;
-
-	signal = dbus_message_new_signal(device->element.path,
-				CONNMAN_DEVICE_INTERFACE, "PropertyChanged");
-	if (signal == NULL)
-		return 0;
-
-	dbus_message_iter_init_append(signal, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	str = policy2string(policy);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-					DBUS_TYPE_STRING_AS_STRING, &value);
-	dbus_message_iter_append_basic(&value, DBUS_TYPE_STRING, &str);
-	dbus_message_iter_close_container(&entry, &value);
-
-	g_dbus_send_message(connection, signal);
-
-	return 0;
-}
-
 static void append_path(gpointer key, gpointer value, gpointer user_data)
 {
 	struct connman_element *element = value;
@@ -376,15 +292,6 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		connman_dbus_dict_append_variant(&dict, "Interface",
 					DBUS_TYPE_STRING, &device->interface);
 
-	str = policy2string(device->policy);
-	if (str != NULL)
-		connman_dbus_dict_append_variant(&dict, "Policy",
-						DBUS_TYPE_STRING, &str);
-
-	if (device->priority > 0)
-		connman_dbus_dict_append_variant(&dict, "Priority",
-					DBUS_TYPE_BYTE, &device->priority);
-
 	connman_dbus_dict_append_variant(&dict, "Powered",
 					DBUS_TYPE_BOOLEAN, &device->powered);
 
@@ -416,6 +323,28 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	return reply;
 }
 
+static gboolean powered_timeout(gpointer user_data)
+{
+	struct connman_device *device = user_data;
+
+	DBG("device %p", device);
+
+	device->timeout = 0;
+
+	if (device->pending != NULL) {
+		DBusMessage *reply;
+
+		reply = __connman_error_operation_timeout(device->pending);
+		if (reply != NULL)
+			g_dbus_send_message(connection, reply);
+
+		dbus_message_unref(device->pending);
+		device->pending = NULL;
+	}
+
+	return FALSE;
+}
+
 static DBusMessage *set_property(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -449,36 +378,23 @@ static DBusMessage *set_property(DBusConnection *conn,
 		dbus_message_iter_get_basic(&value, &powered);
 
 		if (device->powered == powered)
-			return __connman_error_invalid_arguments(msg);
+			return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+
+		if (device->pending != NULL)
+			return __connman_error_in_progress(msg);
 
 		err = set_powered(device, powered);
-		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg, -err);
-	} else if (g_str_equal(name, "Policy") == TRUE) {
-		enum connman_device_policy policy;
-		const char *str;
-		int err;
+		if (err < 0) {
+			if (err != -EINPROGRESS)
+				return __connman_error_failed(msg, -err);
 
-		if (type != DBUS_TYPE_STRING)
-			return __connman_error_invalid_arguments(msg);
+			device->pending = dbus_message_ref(msg);
 
-		dbus_message_iter_get_basic(&value, &str);
-		policy = string2policy(str);
-		if (policy == CONNMAN_DEVICE_POLICY_UNKNOWN)
-			return __connman_error_invalid_arguments(msg);
+			device->timeout = g_timeout_add_seconds(15,
+						powered_timeout, device);
 
-		err = set_policy(conn, device, policy);
-		if (err < 0)
-			return __connman_error_failed(msg, -err);
-	} else if (g_str_equal(name, "Priority") == TRUE) {
-		connman_uint8_t priority;
-
-		if (type != DBUS_TYPE_BYTE)
-			return __connman_error_invalid_arguments(msg);
-
-		dbus_message_iter_get_basic(&value, &priority);
-
-		device->priority = priority;
+			return NULL;
+		}
 	} else if (g_str_equal(name, "ScanInterval") == TRUE) {
 		connman_uint16_t interval;
 
@@ -508,30 +424,59 @@ static DBusMessage *set_property(DBusConnection *conn,
 			device->scan_timeout = g_timeout_add_seconds(interval,
 						device_scan_trigger, device);
 		}
-	} else if (g_str_has_prefix(name, "IPv4") == TRUE) {
+	} else if (g_str_has_prefix(name, "IPv4.") == TRUE) {
+		int err;
+
 		switch (device->mode) {
 		case CONNMAN_DEVICE_MODE_UNKNOWN:
 		case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
 		case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
 			return __connman_error_invalid_arguments(msg);
 		case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
-			__connman_element_set_ipv4(&device->element,
-								name, &value);
 			break;
 		}
-	}
+
+		err = __connman_ipconfig_set_ipv4(device->ipconfig,
+							name + 5, &value);
+		if (err < 0)
+			return __connman_error_failed(msg, -err);
+	} else
+		return __connman_error_invalid_property(msg);
 
 	__connman_storage_save_device(device);
 
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
+static char *build_group(const unsigned char *ssid, unsigned int ssid_len,
+					const char *mode, const char *security)
+{
+	GString *str;
+	unsigned int i;
+
+	str = g_string_sized_new((ssid_len * 2) + 24);
+	if (str == NULL)
+		return NULL;
+
+	if (ssid_len > 0 && ssid[0] != '\0') {
+		for (i = 0; i < ssid_len; i++)
+			g_string_append_printf(str, "%02x", ssid[i]);
+	}
+
+	g_string_append_printf(str, "_%s_%s", mode, security);
+
+	return g_string_free(str, FALSE);
+}
+
 static DBusMessage *join_network(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	struct connman_device *device = data;
 	struct connman_network *network;
 	enum connman_network_type type;
+	unsigned int ssid_size;
+	const char *group, *mode, *security;
+	const void *ssid;
 	DBusMessageIter iter, array;
 	int err, index;
 
@@ -583,6 +528,13 @@ static DBusMessage *join_network(DBusConnection *conn,
 		dbus_message_iter_next(&array);
 	}
 
+	ssid = connman_network_get_blob(network, "WiFi.SSID", &ssid_size);
+	mode = connman_network_get_string(network, "WiFi.Mode");
+	security = connman_network_get_string(network, "WiFi.Security");
+	group = build_group(ssid, ssid_size, mode, security);
+
+	connman_network_set_group(network, group);
+
 	index = connman_device_get_index(device);
 	connman_network_set_index(network, index);
 
@@ -654,7 +606,8 @@ static DBusMessage *propose_scan(DBusConnection *conn,
 
 static GDBusMethodTable device_methods[] = {
 	{ "GetProperties", "",      "a{sv}", get_properties },
-	{ "SetProperty",   "sv",    "",      set_property   },
+	{ "SetProperty",   "sv",    "",      set_property,
+						G_DBUS_METHOD_FLAG_ASYNC },
 	{ "JoinNetwork",   "a{sv}", "",      join_network   },
 	{ "CreateNetwork", "a{sv}", "o",     create_network },
 	{ "RemoveNetwork", "o",     "",      remove_network },
@@ -667,8 +620,6 @@ static GDBusSignalTable device_signals[] = {
 	{ },
 };
 
-static DBusConnection *connection;
-
 static void append_devices(DBusMessageIter *entry)
 {
 	DBusMessageIter value, iter;
@@ -744,10 +695,6 @@ static void device_enable(struct connman_device *device)
 {
 	DBG("device %p", device);
 
-	if (device->policy == CONNMAN_DEVICE_POLICY_IGNORE ||
-				device->policy == CONNMAN_DEVICE_POLICY_OFF)
-		return;
-
 	if (device->powered == TRUE)
 		return;
 
@@ -761,9 +708,6 @@ static void device_disable(struct connman_device *device)
 {
 	DBG("device %p", device);
 
-	if (device->policy == CONNMAN_DEVICE_POLICY_IGNORE)
-		return;
-
 	if (device->powered == FALSE)
 		return;
 
@@ -938,12 +882,24 @@ static void device_destruct(struct connman_element *element)
 
 	DBG("element %p name %s", element, element->name);
 
+	if (device->timeout > 0) {
+		g_source_remove(device->timeout);
+		device->timeout = 0;
+	}
+
+	if (device->pending != NULL) {
+		dbus_message_unref(device->pending);
+		device->pending = NULL;
+	}
+
 	g_free(device->ident);
 	g_free(device->node);
 	g_free(device->name);
 	g_free(device->address);
 	g_free(device->interface);
 
+	connman_ipconfig_unref(device->ipconfig);
+
 	g_free(device->last_network);
 
 	g_hash_table_destroy(device->networks);
@@ -986,46 +942,45 @@ struct connman_device *connman_device_create(const char *node,
 		connman_element_set_string(&device->element,
 					CONNMAN_PROPERTY_ID_TYPE, str);
 
-	device->element.ipv4.method = CONNMAN_IPV4_METHOD_DHCP;
+	device->element.ipv4.method = CONNMAN_IPCONFIG_METHOD_DHCP;
 
 	device->type      = type;
 	device->name      = g_strdup(type2description(device->type));
 	device->mode      = CONNMAN_DEVICE_MODE_UNKNOWN;
-	device->policy    = CONNMAN_DEVICE_POLICY_AUTO;
 	device->secondary = FALSE;
 
 	switch (type) {
 	case CONNMAN_DEVICE_TYPE_UNKNOWN:
 	case CONNMAN_DEVICE_TYPE_VENDOR:
-		device->priority = 0;
 		device->scan_interval = 0;
 		break;
 	case CONNMAN_DEVICE_TYPE_ETHERNET:
 	case CONNMAN_DEVICE_TYPE_WIFI:
-		device->priority = 100;
 		device->scan_interval = 300;
 		break;
 	case CONNMAN_DEVICE_TYPE_WIMAX:
-		device->priority = 20;
 		device->scan_interval = 0;
 		break;
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
-		device->priority = 50;
 		device->scan_interval = 0;
 		break;
 	case CONNMAN_DEVICE_TYPE_GPS:
-		device->priority = 0;
 		device->scan_interval = 0;
 		break;
 	case CONNMAN_DEVICE_TYPE_HSO:
 	case CONNMAN_DEVICE_TYPE_NOZOMI:
 	case CONNMAN_DEVICE_TYPE_HUAWEI:
 	case CONNMAN_DEVICE_TYPE_NOVATEL:
-		device->priority = 60;
 		device->scan_interval = 0;
 		break;
 	}
 
+	device->ipconfig = connman_ipconfig_create();
+	if (device->ipconfig == NULL) {
+		connman_device_unref(device);
+		return NULL;
+	}
+
 	device->networks = g_hash_table_new_full(g_str_hash, g_str_equal,
 						g_free, unregister_network);
 
@@ -1172,19 +1127,6 @@ const char *__connman_device_get_ident(struct connman_device *device)
 	return device->ident;
 }
 
-/**
- * connman_device_set_policy:
- * @device: device structure
- * @policy: power and connection policy
- *
- * Change power and connection policy of device
- */
-void connman_device_set_policy(struct connman_device *device,
-					enum connman_device_policy policy)
-{
-	device->policy = policy;
-}
-
 /**
  * connman_device_set_mode:
  * @device: device structure
@@ -1249,6 +1191,19 @@ int connman_device_set_powered(struct connman_device *device,
 
 	DBG("driver %p powered %d", device, powered);
 
+	if (device->timeout > 0) {
+		g_source_remove(device->timeout);
+		device->timeout = 0;
+	}
+
+	if (device->pending != NULL) {
+		g_dbus_send_reply(connection, device->pending,
+							DBUS_TYPE_INVALID);
+
+		dbus_message_unref(device->pending);
+		device->pending = NULL;
+	}
+
 	if (device->powered == powered)
 		return -EALREADY;
 
@@ -1276,9 +1231,6 @@ int connman_device_set_powered(struct connman_device *device,
 	if (powered == FALSE)
 		return 0;
 
-	if (device->policy != CONNMAN_DEVICE_POLICY_AUTO)
-		return 0;
-
 	if (device->scan_timeout > 0) {
 		g_source_remove(device->scan_timeout);
 		device->scan_timeout = 0;
@@ -1306,7 +1258,7 @@ int connman_device_set_powered(struct connman_device *device,
 int connman_device_set_carrier(struct connman_device *device,
 						connman_bool_t carrier)
 {
-	DBG("driver %p carrier %d", device, carrier);
+	DBG("device %p carrier %d", device, carrier);
 
 	switch (device->mode) {
 	case CONNMAN_DEVICE_MODE_UNKNOWN:
@@ -1332,6 +1284,9 @@ int __connman_device_connect(struct connman_device *device)
 	if (device->disconnected == FALSE)
 		return -EINVAL;
 
+	if (device->driver && device->driver->connect)
+		device->driver->connect(device);
+
 	return 0;
 }
 
@@ -1353,6 +1308,9 @@ int __connman_device_disconnect(struct connman_device *device)
 			__connman_network_disconnect(network);
 	}
 
+	if (device->driver && device->driver->disconnect)
+		device->driver->disconnect(device);
+
 	return 0;
 }
 
@@ -1361,6 +1319,7 @@ static void connect_known_network(struct connman_device *device)
 	struct connman_network *network = NULL;
 	GHashTableIter iter;
 	gpointer key, value;
+	const char *name;
 	unsigned int count = 0;
 
 	DBG("device %p", device);
@@ -1368,9 +1327,7 @@ static void connect_known_network(struct connman_device *device)
 	g_hash_table_iter_init(&iter, device->networks);
 
 	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
-		connman_uint8_t old_priority, new_priority;
 		connman_uint8_t old_strength, new_strength;
-		const char *name;
 
 		count++;
 
@@ -1386,25 +1343,11 @@ static void connect_known_network(struct connman_device *device)
 			}
 		}
 
-		if (connman_network_get_remember(value) == FALSE)
-			continue;
-
 		if (network == NULL) {
 			network = value;
 			continue;
 		}
 
-		old_priority = connman_network_get_uint8(network,
-						CONNMAN_PROPERTY_ID_PRIORITY);
-		new_priority = connman_network_get_uint8(value,
-						CONNMAN_PROPERTY_ID_PRIORITY);
-
-		if (new_priority != old_priority) {
-			if (new_priority > old_priority)
-				network = value;
-			continue;
-		}
-
 		old_strength = connman_network_get_uint8(network,
 						CONNMAN_PROPERTY_ID_STRENGTH);
 		new_strength = connman_network_get_uint8(value,
@@ -1417,9 +1360,13 @@ static void connect_known_network(struct connman_device *device)
 	if (network != NULL) {
 		int err;
 
-		err = connman_network_connect(network);
-		if (err == 0 || err == -EINPROGRESS)
-			return;
+		name = connman_network_get_string(value,
+						CONNMAN_PROPERTY_ID_NAME);
+		if (name != NULL) {
+			err = __connman_network_connect(network);
+			if (err == 0 || err == -EINPROGRESS)
+				return;
+		}
 	}
 
 	if (count > 0)
@@ -1448,9 +1395,6 @@ static gboolean remove_unavailable_network(gpointer key, gpointer value,
 	if (connman_network_get_connected(network) == TRUE)
 		return FALSE;
 
-	if (connman_network_get_remember(network) == TRUE)
-		return FALSE;
-
 	if (connman_network_get_available(network) == TRUE)
 		return FALSE;
 
@@ -1471,7 +1415,7 @@ int connman_device_set_scanning(struct connman_device *device,
 	DBusMessageIter entry, value;
 	const char *key = "Scanning";
 
-	DBG("driver %p scanning %d", device, scanning);
+	DBG("device %p scanning %d", device, scanning);
 
 	if (!device->driver || !device->driver->scan)
 		return -EINVAL;
@@ -1523,9 +1467,6 @@ int connman_device_set_scanning(struct connman_device *device,
 	if (device->disconnected == TRUE)
 		return 0;
 
-	if (device->policy != CONNMAN_DEVICE_POLICY_AUTO)
-		return 0;
-
 	connect_known_network(device);
 
 	return 0;
@@ -1541,7 +1482,7 @@ int connman_device_set_scanning(struct connman_device *device,
 int connman_device_set_disconnected(struct connman_device *device,
 						connman_bool_t disconnected)
 {
-	DBG("driver %p disconnected %d", device, disconnected);
+	DBG("device %p disconnected %d", device, disconnected);
 
 	switch (device->mode) {
 	case CONNMAN_DEVICE_MODE_UNKNOWN:
@@ -1560,6 +1501,33 @@ int connman_device_set_disconnected(struct connman_device *device,
 	return 0;
 }
 
+/**
+ * connman_device_set_connected:
+ * @device: device structure
+ * @connected: connected state
+ *
+ * Change connected state of device (for Ethernet like devices)
+ */
+int connman_device_set_connected(struct connman_device *device,
+						connman_bool_t connected)
+{
+	DBG("device %p connected %d", device, connected);
+
+	switch (device->mode) {
+	case CONNMAN_DEVICE_MODE_UNKNOWN:
+	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		return -EINVAL;
+	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		break;
+	}
+
+	if (device->carrier == FALSE)
+		return -ENOTCONN;
+
+	return set_connected(device, connected);
+}
+
 /**
  * connman_device_set_string:
  * @device: device structure
@@ -1728,13 +1696,25 @@ void __connman_device_set_network(struct connman_device *device,
 {
 	const char *name;
 
+	if (device->network == network)
+		return;
+
+	if (device->network != NULL)
+		connman_network_unref(device->network);
+
 	if (network != NULL) {
 		name = connman_network_get_string(network,
 						CONNMAN_PROPERTY_ID_NAME);
+		g_free(device->last_network);
 		device->last_network = g_strdup(name);
-	}
 
-	device->network = network;
+		device->network = connman_network_ref(network);
+	} else {
+		g_free(device->last_network);
+		device->last_network = NULL;
+
+		device->network = NULL;
+	}
 }
 
 /**
@@ -1865,15 +1845,14 @@ static struct connman_driver device_driver = {
 static int device_load(struct connman_device *device)
 {
 	GKeyFile *keyfile;
-	gchar *pathname, *data = NULL;
+	gchar *pathname, *identifier, *data = NULL;
 	gsize length;
-	char *str;
 	int val;
 
 	DBG("device %p", device);
 
 	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR,
-							device->element.name);
+					__connman_profile_active_ident());
 	if (pathname == NULL)
 		return -ENOMEM;
 
@@ -1894,16 +1873,9 @@ static int device_load(struct connman_device *device)
 
 	g_free(data);
 
-	str = g_key_file_get_string(keyfile, "Configuration", "Policy", NULL);
-	if (str != NULL) {
-		device->policy = string2policy(str);
-		g_free(str);
-	}
-
-	val = g_key_file_get_integer(keyfile, "Configuration",
-							"Priority", NULL);
-	if (val > 0)
-		device->priority = val;
+	identifier = g_strdup_printf("device_%s", device->element.name);
+	if (identifier == NULL)
+		goto done;
 
 	switch (device->mode) {
 	case CONNMAN_DEVICE_MODE_UNKNOWN:
@@ -1911,36 +1883,31 @@ static int device_load(struct connman_device *device)
 		break;
 	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
 	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
-		val = g_key_file_get_integer(keyfile, "Configuration",
+		val = g_key_file_get_integer(keyfile, identifier,
 							"ScanInterval", NULL);
 		if (val > 0)
 			device->scan_interval = val;
 		break;
 	}
 
-#if 0
-	str = g_key_file_get_string(keyfile, "Configuration",
-							"LastNetwork", NULL);
-	if (str != NULL)
-		device->last_network = str;
-#endif
-
+done:
 	g_key_file_free(keyfile);
 
+	g_free(identifier);
+
 	return 0;
 }
 
 static int device_save(struct connman_device *device)
 {
 	GKeyFile *keyfile;
-	gchar *pathname, *data = NULL;
+	gchar *pathname, *identifier = NULL, *data = NULL;
 	gsize length;
-	const char *str;
 
 	DBG("device %p", device);
 
 	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR,
-							device->element.name);
+					__connman_profile_active_ident());
 	if (pathname == NULL)
 		return -ENOMEM;
 
@@ -1958,13 +1925,9 @@ static int device_save(struct connman_device *device)
 	g_free(data);
 
 update:
-	str = policy2string(device->policy);
-	if (str != NULL)
-		g_key_file_set_string(keyfile, "Configuration", "Policy", str);
-
-	if (device->priority > 0)
-		g_key_file_set_integer(keyfile, "Configuration",
-						"Priority", device->priority);
+	identifier = g_strdup_printf("device_%s", device->element.name);
+	if (identifier == NULL)
+		goto done;
 
 	switch (device->mode) {
 	case CONNMAN_DEVICE_MODE_UNKNOWN:
@@ -1973,15 +1936,11 @@ update:
 	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
 	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
 		if (device->scan_interval > 0)
-			g_key_file_set_integer(keyfile, "Configuration",
+			g_key_file_set_integer(keyfile, identifier,
 					"ScanInterval", device->scan_interval);
 		break;
 	}
 
-	if (device->last_network != NULL)
-		g_key_file_set_string(keyfile, "Configuration",
-					"LastNetwork", device->last_network);
-
 	data = g_key_file_to_data(keyfile, &length, NULL);
 
 	if (g_file_set_contents(pathname, data, length, NULL) == FALSE)
@@ -1992,11 +1951,9 @@ done:
 
 	g_key_file_free(keyfile);
 
+	g_free(identifier);
 	g_free(pathname);
 
-	if (device->network != NULL)
-		__connman_storage_save_network(device->network);
-
 	return 0;
 }
 
